/*
 * Copyright 2022 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wNewmark.h"
#include "wMshData.h"
#include "wNode.h"
#include "wElement.h"
#include "wTimer.h"
#include "wTag.h"
#include "wProblem.h"
#include "wDisplayHook.h"
#include "wSource.h"
#include "wResults.h"
#include "wMshExport.h"
using namespace tbox;
using namespace waves;

Newmark::Newmark(std::shared_ptr<waves::Problem> _pbl) : TimeIntegration(_pbl)
{
    gamma = 0.5;
    beta = 0.25;
}

void Newmark::start(MshExport *mshWriter)
{
    std::cout << *this;
    chrono1.start();

    auto msh = pbl->msh;
    u1.resize(msh->nodes.size());
    v1.resize(msh->nodes.size());
    Eigen::Map<Eigen::VectorXd> u1_(u1.data(), u1.size()), v1_(v1.data(), v1.size());

    // matrix assembly
    Eigen::SparseMatrix<double, Eigen::RowMajor> K(msh->nodes.size(), msh->nodes.size());
    std::vector<double> Md(msh->nodes.size());
    Eigen::Map<Eigen::VectorXd> Md_(Md.data(), Md.size());
    buildKM(K, Md, u1);

    Eigen::SparseMatrix<double, Eigen::RowMajor> S(msh->nodes.size(), msh->nodes.size());
    buildS(S);

    // setup results
    Results results;
    results.scalars_at_nodes["u"] = &u1;
    // results.scalars_at_nodes["v"] = &v1;

    // time integration

    std::cout << "\n\nTime integration...\n";
    std::cout << *this << '\n';

    double t = 0.0;
    int nt = 1;
    results.nt = nt;
    results.time = t;

    while (t < ttot && !stopit)
    {
        if (verbose)
            std::cout << "nt " << nt << " t=" << t << "\n";

        for (auto s : pbl->srcs)
            s->apply(t, u1);

        t += dt;
        nt++;

        // equation #1
        u1_ += v1_ * dt; // u1 = u0 + dt*v0

        for (auto s : pbl->srcs)
            s->apply(t, u1);

        // equation #2
        v1_.array() -= dt * (K * u1_ + S * v1_).array() / Md_.array();

        if (nt % savefreq == 0)
        {
            // std::cout << "nt " << nt << " t=" << t << "\n";
            results.time = t;
            results.nt++;
            // mshWriter->save("res", results);

            if (dhook)
                dhook->display(nt, t, u1);
        }
        if (dhook)
            dhook->refresh();
    }

    chrono1.stop();
    std::cout << "[cpu] " << chrono1 << '\n';
}
